@@ -1,9 +1,78 @@
 use _007::Q;
-use _007::Parser::Exceptions;
+use _007::Parser::Syntax;
+
+class X::String::Newline is Exception {
+    method message { "Found a newline inside a string literal" }
+}
+
+class X::PointyBlock::SinkContext is Exception {
+    method message { "Pointy blocks cannot occur on the statement level" }
+}
+
+class X::Trait::Conflict is Exception {
+    has Str $.trait1;
+    has Str $.trait2;
+
+    method message { "Traits '$.trait1' and '$.trait2' cannot coexist on the same routine" }
+}
+
+class X::Trait::Duplicate is Exception {
+    has Str $.trait;
+
+    method message { "Trait '$.trait' is used more than once" }
+}
+
+class X::Trait::IllegalValue is Exception {
+    has Str $.trait;
+    has Str $.value;
+
+    method message { "The value '$.value' is not compatible with the trait '$.trait'" }
+}
+
+class X::Macro::Postdeclared is Exception {
+    has Str $.name;
+
+    method message { "Macro $.name declared after it was called" }
+}
+
+class X::Op::Nonassociative is Exception {
+    has Str $.op1;
+    has Str $.op2;
+
+    method message {
+        my $name1 = $.op1.type.substr(1, *-1);
+        my $name2 = $.op2.type.substr(1, *-1);
+        "'$name1' and '$name2' do not associate -- please use parentheses"
+    }
+}
+
+class X::Precedence::Incompatible is Exception {
+    method message { "Trying to relate a pre/postfix operator with an infix operator" }
+}
+
+class X::Property::NotDeclared is Exception {
+    has Str $.type;
+    has Str $.property;
+
+    method message { "The property '$.property' is not defined on type '$.type'" }
+}
+
+class X::Property::Required is Exception {
+    has Str $.type;
+    has Str $.property;
+
+    method message { "The property '$.property' is required on type '$.type'" }
+}
+
+class X::Property::Duplicate is Exception {
+    has Str $.property;
+
+    method message { "The property '$.property' was declared more than once in a property list" }
+}
 
 class _007::Parser::Actions {
     method finish-block($block) {
-        $block.static-lexpad = $*runtime.current-frame.pad;
+        $block.static-lexpad = $*runtime.current-frame.properties<pad>;
         $*runtime.leave;
     }
 
@@ -27,7 +96,7 @@ class _007::Parser::Actions {
     method statement:my ($/) {
         make Q::Statement::My.new(
             :identifier($<identifier>.ast),
-            :expr($<EXPR> ?? $<EXPR>.ast !! Val::None.new));
+            :expr($<EXPR> ?? $<EXPR>.ast !! NONE));
     }
 
     method statement:constant ($/) {
@@ -38,9 +107,8 @@ class _007::Parser::Actions {
             :identifier($<identifier>.ast),
             :expr($<EXPR>.ast));
 
-        my $name = $<identifier>.ast.name.value;
         my $value = $<EXPR>.ast.eval($*runtime);
-        $*runtime.put-var($name, $value);
+        $<identifier>.ast.put-value($value, $*runtime);
     }
 
     method statement:expr ($/) {
@@ -49,7 +117,8 @@ class _007::Parser::Actions {
         #      in the expression tree
         if $<EXPR>.ast ~~ Q::Block {
             make Q::Statement::Expr.new(:expr(Q::Postfix::Call.new(
-                :operand($<EXPR>.ast),
+                :identifier(Q::Identifier.new(:name(Val::Str.new(:value("postfix:()"))))),
+                :operand(Q::Term::Sub.new(:identifier(NONE), :block($<EXPR>.ast))),
                 :argumentlist(Q::ArgumentList.new)
             )));
         }
@@ -66,8 +135,8 @@ class _007::Parser::Actions {
 
     sub maybe-install-operator($identname, @trait) {
         return
-            unless $identname ~~ / (< prefix infix postfix >)
-                                    ':<' (<-[>]>+) '>' /;
+            unless $identname ~~ /^ (< prefix infix postfix >)
+                                    ':' (.+) /;
 
         my $type = ~$0;
         my $op = ~$1;
@@ -84,11 +153,11 @@ class _007::Parser::Actions {
                     unless $identifier ~~ Q::Identifier;
                 sub check-if-op($s) {
                     die "Unknown thing in '$name' trait"
-                        unless $s ~~ /< pre in post > 'fix:<' (<-[>]>+) '>'/;
-                    %precedence{$name} = ~$0;
+                        unless $s ~~ /^ < pre in post > 'fix:' /;
                     die X::Precedence::Incompatible.new
                         if $type eq ('prefix' | 'postfix') && $s ~~ /^ in/
                         || $type eq 'infix' && $s ~~ /^ < pre post >/;
+                    %precedence{$name} = $s;
                 }($identifier.name);
             }
             elsif $name eq "assoc" {
@@ -111,48 +180,54 @@ class _007::Parser::Actions {
                 if %precedence{$trait1} && %precedence{$trait2};
         }
 
-        $*parser.oplevel.install($type, $op, :%precedence, :$assoc);
+        $*parser.opscope.install($type, $op, :%precedence, :$assoc);
     }
 
     method statement:sub-or-macro ($/) {
         my $identifier = $<identifier>.ast;
-        my $name = ~$<identifier>;
+        my $name = $<identifier>.ast.name;
         my $parameterlist = $<parameterlist>.ast;
         my $traitlist = $<traitlist>.ast;
         my $statementlist = $<blockoid>.ast;
 
         my $block = Q::Block.new(:$parameterlist, :$statementlist);
-        my %static-lexpad = $*runtime.current-frame.pad;
+        my $static-lexpad = $*runtime.current-frame.properties<pad>;
         self.finish-block($block);
 
         my $outer-frame = $*runtime.current-frame;
         my $val;
         if $<routine> eq "sub" {
             make Q::Statement::Sub.new(:$identifier, :$traitlist, :$block);
-            $val = Val::Sub.new(:$name, :$parameterlist, :$statementlist, :$outer-frame, :%static-lexpad);
+            $val = Val::Sub.new(:$name, :$parameterlist, :$statementlist, :$outer-frame, :$static-lexpad);
         }
         elsif $<routine> eq "macro" {
             make Q::Statement::Macro.new(:$identifier, :$traitlist, :$block);
-            $val = Val::Macro.new(:$name, :$parameterlist, :$statementlist, :$outer-frame, :%static-lexpad);
+            $val = Val::Macro.new(:$name, :$parameterlist, :$statementlist, :$outer-frame, :$static-lexpad);
         }
         else {
             die "Unknown routine type $<routine>"; # XXX: Turn this into an X:: exception
         }
 
-        $*runtime.put-var($name, $val);
+        $identifier.put-value($val, $*runtime);
 
         maybe-install-operator($name, $<traitlist><trait>);
     }
 
     method statement:return ($/) {
-        make Q::Statement::Return.new(:expr($<EXPR> ?? $<EXPR>.ast !! Val::None.new));
+        die X::ControlFlow::Return.new
+            unless $*insub;
+        make Q::Statement::Return.new(:expr($<EXPR> ?? $<EXPR>.ast !! NONE));
+    }
+
+    method statement:throw ($/) {
+        make Q::Statement::Throw.new(:expr($<EXPR> ?? $<EXPR>.ast !! NONE));
     }
 
     method statement:if ($/) {
         my %parameters = $<xblock>.ast;
         %parameters<else> = $<else> :exists
             ?? $<else>.ast
-            !! Val::None.new;
+            !! NONE;
 
         make Q::Statement::If.new(|%parameters);
     }
@@ -171,6 +246,17 @@ class _007::Parser::Actions {
         $*runtime.run(Q::CompUnit.new(:$block));
     }
 
+    method statement:class ($/) {
+        my $identifier = $<identifier>.ast;
+        my $block = $<block>.ast;
+        make Q::Statement::Class.new(:$block);
+        my $val = Val::Type.of(class :: {
+            method attributes { () }
+            method ^name($) { $identifier.name.value }
+        });
+        $identifier.put-value($val, $*runtime);
+    }
+
     method traitlist($/) {
         my @traits = $<trait>».ast;
         if bag( @traits.map: *.identifier.name.value ).grep( *.value > 1 )[0] -> $p {
@@ -195,9 +281,11 @@ class _007::Parser::Actions {
     }
     method pblock ($/) {
         if $<parameterlist> {
-            make Q::Block.new(
+            my $block = Q::Block.new(
                 :parameterlist($<parameterlist>.ast),
                 :statementlist($<blockoid>.ast));
+            make $block;
+            self.finish-block($block);
         } else {
             make $<block>.ast;
         }
@@ -211,23 +299,23 @@ class _007::Parser::Actions {
 
     method EXPR($/) {
         sub name($op) {
-            $op.identifier.name.value.subst(/^ \w+ ":<"/, "").subst(/">" $/, "");
+            $op.identifier.name.value;
         }
 
-        sub tighter($op1, $op2, $_ = $*parser.oplevel.infixprec) {
+        sub tighter($op1, $op2, $_ = $*parser.opscope.infixprec) {
             .first(*.contains(name($op1)), :k) > .first(*.contains(name($op2)), :k);
         }
 
-        sub equal($op1, $op2, $_ = $*parser.oplevel.infixprec) {
+        sub equal($op1, $op2, $_ = $*parser.opscope.infixprec) {
             .first(*.contains(name($op1)), :k) == .first(*.contains(name($op2)), :k);
         }
 
         sub left-associative($op) {
-            return $*parser.oplevel.infixprec.first(*.contains(name($op))).assoc eq "left";
+            return $*parser.opscope.infixprec.first(*.contains(name($op))).assoc eq "left";
         }
 
         sub non-associative($op) {
-            return $*parser.oplevel.infixprec.first(*.contains(name($op))).assoc eq "non";
+            return $*parser.opscope.infixprec.first(*.contains(name($op))).assoc eq "non";
         }
 
         my @opstack;
@@ -244,15 +332,19 @@ class _007::Parser::Actions {
 
             my $c = $*runtime.maybe-get-var($infix.identifier.name.value);
             if $c ~~ Val::Macro {
-                @termstack.push($*runtime.call($c, [$t1, $t2]));
+                my $expansion = $*runtime.call($c, [$t1, $t2]);
+                if $*unexpanded {
+                    @termstack.push($infix.new(:lhs($t1), :rhs($t2), :identifier($infix.identifier)));
+                }
+                else {
+                    @termstack.push($expansion);
+                }
             }
             else {
                 @termstack.push($infix.new(:lhs($t1), :rhs($t2), :identifier($infix.identifier)));
 
-                if $infix ~~ Q::Infix::Assignment {
-                    die X::Immutable.new(:method<assignment>, :typename($t1.^name))
-                        unless $t1 ~~ Q::Identifier;
-                    my $block = $*runtime.current-frame();
+                if $infix ~~ Q::Infix::Assignment && $t1 ~~ Q::Identifier {
+                    my $frame = $*runtime.current-frame;
                     my $symbol = $t1.name.value;
                     die X::Undeclared.new(:$symbol)
                         unless @*declstack[*-1]{$symbol} :exists;
@@ -260,7 +352,7 @@ class _007::Parser::Actions {
                     my $declname = $decltype.^name.subst(/ .* '::'/, "").lc;
                     die X::Assignment::RO.new(:typename("$declname '$symbol'"))
                         unless $decltype.is-assignable;
-                    %*assigned{$block ~ $symbol}++;
+                    %*assigned{$frame.id ~ $symbol}++;
                 }
             }
         }
@@ -284,23 +376,23 @@ class _007::Parser::Actions {
 
     method termish($/) {
         sub name($op) {
-            $op.identifier.name.value.subst(/^ \w+ ":<"/, "").subst(/">" $/, "");
+            $op.identifier.name.value;
         }
 
-        sub tighter($op1, $op2, $_ = $*parser.oplevel.prepostfixprec) {
+        sub tighter($op1, $op2, $_ = $*parser.opscope.prepostfixprec) {
             .first(*.contains(name($op1)), :k) > .first(*.contains(name($op2)), :k);
         }
 
-        sub equal($op1, $op2, $_ = $*parser.oplevel.prepostfixprec) {
+        sub equal($op1, $op2, $_ = $*parser.opscope.prepostfixprec) {
             .first(*.contains(name($op1)), :k) == .first(*.contains(name($op2)), :k);
         }
 
         sub left-associative($op) {
-            return $*parser.oplevel.prepostfixprec.first(*.contains(name($op))).assoc eq "left";
+            return $*parser.opscope.prepostfixprec.first(*.contains(name($op))).assoc eq "left";
         }
 
         sub non-associative($op) {
-            return $*parser.oplevel.prepostfixprec.first(*.contains(name($op))).assoc eq "non";
+            return $*parser.opscope.prepostfixprec.first(*.contains(name($op))).assoc eq "non";
         }
 
         make $<term>.ast;
@@ -318,7 +410,13 @@ class _007::Parser::Actions {
 
             my $c = $*runtime.maybe-get-var($prefix.identifier.name.value);
             if $c ~~ Val::Macro {
-                make $*runtime.call($c, [$/.ast]);
+                my $expansion = $*runtime.call($c, [$/.ast]);
+                if $*unexpanded {
+                    make $prefix.new(:operand($/.ast), :identifier($prefix.identifier));
+                }
+                else {
+                    make $expansion;
+                }
             }
             else {
                 make $prefix.new(:operand($/.ast), :identifier($prefix.identifier));
@@ -333,8 +431,25 @@ class _007::Parser::Actions {
             && $/.ast ~~ Q::Identifier
             && (my $macro = $*runtime.maybe-get-var($/.ast.name.value)) ~~ Val::Macro {
                 my @arguments = $postfix.argumentlist.arguments.elements;
-                my $qtree = $*runtime.call($macro, @arguments);
-                make $qtree;
+                my $expansion = $*runtime.call($macro, @arguments);
+
+                if $expansion ~~ Q::Statement::My {
+                    _007::Parser::Syntax::declare(Q::Statement::My, $expansion.identifier.name.value);
+                }
+
+                if $*unexpanded {
+                    make $postfix.new(:$identifier, :operand($/.ast), :argumentlist($postfix.argumentlist));
+                }
+                elsif $expansion ~~ Q::Statement {
+                    make Q::Expr::StatementListAdapter.new(
+                        :statementlist(Q::StatementList.new(
+                            :statements(Val::Array.new(:elements([$expansion])))
+                        ))
+                    );
+                }
+                else {
+                    make $expansion;
+                }
             }
             elsif $postfix ~~ Q::Postfix::Index {
                 make $postfix.new(:$identifier, :operand($/.ast), :index($postfix.index));
@@ -348,7 +463,13 @@ class _007::Parser::Actions {
             else {
                 my $c = $*runtime.maybe-get-var($postfix.identifier.name.value);
                 if $c ~~ Val::Macro {
-                    make $*runtime.call($c, [$/.ast]);
+                    my $expansion = $*runtime.call($c, [$/.ast]);
+                    if $*unexpanded {
+                        make $postfix.new(:$identifier, :operand($/.ast));
+                    }
+                    else {
+                        make $expansion;
+                    }
                 }
                 else {
                     make $postfix.new(:$identifier, :operand($/.ast));
@@ -366,7 +487,7 @@ class _007::Parser::Actions {
             else {
                 my $prefix = @prefixes[0].ast;
                 my $postfix = @postfixes[0].ast;
-                die X::Op::Nonassociative.new(:op1($prefix.identifier.name.value), :op2($postfix.identifier.name.value))
+                die X::Op::Nonassociative.new(:op1(name($prefix)), :op2(name($postfix)))
                     if equal($prefix, $postfix) && non-associative($prefix);
                 if tighter($prefix, $postfix)
                     || equal($prefix, $postfix) && left-associative($prefix) {
@@ -383,10 +504,10 @@ class _007::Parser::Actions {
     method prefix($/) {
         my $op = ~$/;
         my $identifier = Q::Identifier.new(
-            :name(Val::Str.new(:value("prefix:<$op>"))),
+            :name(Val::Str.new(:value("prefix:$op"))),
             :frame($*runtime.current-frame),
         );
-        make $*parser.oplevel.ops<prefix>{$op}.new(:$identifier, :operand(Val::None));
+        make $*parser.opscope.ops<prefix>{$op}.new(:$identifier, :operand(Val::NoneType));
     }
 
     method prefix-unquote($/) {
@@ -407,6 +528,14 @@ class _007::Parser::Actions {
         make Q::Literal::None.new;
     }
 
+    method term:false ($/) {
+        make Q::Literal::Bool.new(:value(Val::Bool.new(:value(False))));
+    }
+
+    method term:true ($/) {
+        make Q::Literal::Bool.new(:value(Val::Bool.new(:value(True))));
+    }
+
     method term:int ($/) {
         make Q::Literal::Int.new(:value(Val::Int.new(:value(+$/))));
     }
@@ -423,8 +552,23 @@ class _007::Parser::Actions {
         make $<EXPR>.ast;
     }
 
+    method term:regex ($/) {
+        make Q::Term::Regex.new(:contents($<contents>.ast.value));
+    }
+
     method term:identifier ($/) {
         make $<identifier>.ast;
+        my $name = $<identifier>.ast.name.value;
+        if !$*runtime.declared($name) {
+            my $frame = $*runtime.current-frame;
+            $*parser.postpone: sub checking-postdeclared {
+                my $value = $*runtime.maybe-get-var($name, $frame);
+                die X::Macro::Postdeclared.new(:$name)
+                    if $value ~~ Val::Macro;
+                die X::Undeclared.new(:symbol($name))
+                    unless $value ~~ Val::Sub;
+            };
+        }
     }
 
     method term:block ($/) {
@@ -432,12 +576,48 @@ class _007::Parser::Actions {
     }
 
     method term:quasi ($/) {
+        my $qtype = Val::Str.new(:value(~($<qtype> // "")));
+
+        if $<block> -> $block {
+            # If the quasi consists of a block with a single expression statement, it's very
+            # likely that what we want to inject is the expression, not the block.
+            #
+            # The exception to that heuristic is when we've explicitly specified `@ Q::Block`
+            # on the quasi.
+            #
+            # This is not a "nice" solution, nor a comprehensive one. In the end it's connected
+            # to the troubled musings in <https://github.com/masak/007/issues/7>, which aren't
+            # completely solved yet.
+
+            if $qtype.value eq "Q::Statement" {
+                # XXX: make sure there's only one statement (suboptimal; should parse-error sooner)
+                my $contents = $block.ast.statementlist.statements.elements[0];
+                make Q::Term::Quasi.new(:$contents, :$qtype);
+                return;
+            }
+            elsif $qtype.value eq "Q::StatementList" {
+                my $contents = $block.ast.statementlist;
+                make Q::Term::Quasi.new(:$contents, :$qtype);
+                return;
+            }
+            elsif $qtype.value ne "Q::Block"
+                && $block.ast ~~ Q::Block
+                && $block.ast.statementlist.statements.elements.elems == 1
+                && $block.ast.statementlist.statements.elements[0] ~~ Q::Statement::Expr {
+
+                my $contents = $block.ast.statementlist.statements.elements[0].expr;
+                make Q::Term::Quasi.new(:$contents, :$qtype);
+                return;
+            }
+        }
+
         for <argumentlist block compunit EXPR infix parameter parameterlist
             postfix prefix property propertylist statement statementlist
             term trait traitlist unquote> -> $subrule {
 
             if $/{$subrule} -> $submatch {
-                make Q::Term::Quasi.new(:contents($submatch.ast));
+                my $contents = $submatch.ast;
+                make Q::Term::Quasi.new(:$contents, :$qtype);
                 return;
             }
         }
@@ -452,17 +632,18 @@ class _007::Parser::Actions {
 
         my $block = Q::Block.new(:$parameterlist, :$statementlist);
         if $<identifier> {
-            my $name = ~$<identifier>;
-            my $outer-frame = $*runtime.current-frame;  # XXX: this is not really the outer frame, is it?
-            my %static-lexpad = $*runtime.current-frame.pad;
-            my $val = Val::Sub.new(:$name, :$parameterlist, :$statementlist, :$outer-frame, :%static-lexpad);
-            $*runtime.put-var($name, $val);
+            my $name = $<identifier>.ast.name;
+            my $outer-frame = $*runtime.current-frame.properties<block>.outer-frame;
+            my $static-lexpad = $*runtime.current-frame.properties<pad>;
+            my $val = Val::Sub.new(:$name, :$parameterlist, :$statementlist, :$outer-frame, :$static-lexpad);
+            $<identifier>.ast.put-value($val, $*runtime);
         }
         self.finish-block($block);
 
+        my $name = $<identifier>.ast.name;
         my $identifier = $<identifier>
-            ?? Q::Identifier.new(:name(Val::Str.new(:value(~$<identifier>))))
-            !! Val::None.new;
+            ?? Q::Identifier.new(:$name)
+            !! NONE;
         make Q::Term::Sub.new(:$identifier, :$traitlist, :$block);
     }
 
@@ -470,8 +651,8 @@ class _007::Parser::Actions {
         make Q::Unquote.new(:expr($<EXPR>.ast));
     }
 
-    method term:object ($/) {
-        my $type = ~($<identifier> // "Object");
+    method term:new-object ($/) {
+        my $type = $<identifier>.ast.name.value;
         my $type-obj = $*runtime.get-var($type).type;
 
         if $type-obj !=== Val::Object {
@@ -497,6 +678,15 @@ class _007::Parser::Actions {
             :propertylist($<propertylist>.ast));
     }
 
+    method term:object ($/) {
+        my $type = "Object";
+        my $name = Val::Str.new(:value($type));
+
+        make Q::Term::Object.new(
+            :type(Q::Identifier.new(:$name)),
+            :propertylist($<propertylist>.ast));
+    }
+
     method propertylist ($/) {
         my %seen;
         for $<property>».ast -> Q::Property $p {
@@ -513,28 +703,33 @@ class _007::Parser::Actions {
     }
 
     method property:identifier-expr ($/) {
-        make Q::Property.new(:key(Val::Str.new(:value(~$<identifier>))), :value($<value>.ast));
+        my $key = $<identifier>.ast.name;
+        make Q::Property.new(:$key, :value($<value>.ast));
     }
 
     method property:identifier ($/) {
-        make Q::Property.new(:key(Val::Str.new(:value(~$<identifier>))), :value($<identifier>.ast));
+        my $key = $<identifier>.ast.name;
+        make Q::Property.new(:$key, :value($<identifier>.ast));
     }
 
     method property:method ($/) {
-        make Q::Property.new(:key(Val::Str.new(:value(~$<identifier>))), :value(Q::Term::Sub.new(
-            :identifier(Q::Identifier.new(:name(Val::Str.new(:value(~$<identifier>))))),
-            :block(Q::Block.new(
-                :parameterlist($<parameterlist>.ast),
-                :statementlist($<blockoid>.ast))))));
+        my $block = Q::Block.new(
+            :parameterlist($<parameterlist>.ast),
+            :statementlist($<blockoid>.ast));
+        my $name = $<identifier>.ast.name;
+        my $identifier = Q::Identifier.new(:$name);
+        make Q::Property.new(:key($name), :value(
+            Q::Term::Sub.new(:$identifier, :$block)));
+        self.finish-block($block);
     }
 
     method infix($/) {
         my $op = ~$/;
         my $identifier = Q::Identifier.new(
-            :name(Val::Str.new(:value("infix:<$op>"))),
+            :name(Val::Str.new(:value("infix:$op"))),
             :frame($*runtime.current-frame),
         );
-        make $*parser.oplevel.ops<infix>{$op}.new(:$identifier, :lhs(Val::None.new), :rhs(Val::None.new));
+        make $*parser.opscope.ops<infix>{$op}.new(:$identifier, :lhs(NONE), :rhs(NONE));
     }
 
     method infix-unquote($/) {
@@ -557,27 +752,35 @@ class _007::Parser::Actions {
             $op = ".";
         }
         my $identifier = Q::Identifier.new(
-            :name(Val::Str.new(:value("postfix:<$op>"))),
+            :name(Val::Str.new(:value("postfix:$op"))),
             :frame($*runtime.current-frame),
         );
         # XXX: this can't stay hardcoded forever, but we don't have the machinery yet
         # to do these right enough
         if $<index> {
-            make Q::Postfix::Index.new(index => $<EXPR>.ast, :$identifier, :operand(Val::None.new));
+            make Q::Postfix::Index.new(index => $<EXPR>.ast, :$identifier, :operand(NONE));
         }
         elsif $<call> {
-            make Q::Postfix::Call.new(argumentlist => $<argumentlist>.ast, :$identifier, :operand(Val::None.new));
+            make Q::Postfix::Call.new(argumentlist => $<argumentlist>.ast, :$identifier, :operand(NONE));
         }
         elsif $<prop> {
-            make Q::Postfix::Property.new(property => $<identifier>.ast, :$identifier, :operand(Val::None.new));
+            make Q::Postfix::Property.new(property => $<identifier>.ast, :$identifier, :operand(NONE));
         }
         else {
-            make $*parser.oplevel.ops<postfix>{$op}.new(:$identifier, :operand(Val::None.new));
+            make $*parser.opscope.ops<postfix>{$op}.new(:$identifier, :operand(NONE));
         }
     }
 
     method identifier($/) {
-        make Q::Identifier.new(:name(Val::Str.new(:value(~$/))));
+        my $value = ~$/;
+        sub () {
+            $value ~~ s[':<' ([ '\\>' | '\\\\' | <-[>]> ]+) '>' $] = ":{$0}";
+            $value ~~ s[':«' ([ '\\»' | '\\\\' | <-[»]> ]+) '»' $] = ":{$0}";
+            $value ~~ s:g['\\>'] = '>';
+            $value ~~ s:g['\\»'] = '»';
+            $value ~~ s:g['\\\\'] = '\\';
+        }();
+        make Q::Identifier.new(:name(Val::Str.new(:$value)));
     }
 
     method argumentlist($/) {
